from typing import Optional

from matplotlib.patches import Circle, Wedge
from matplotlib.pyplot import subplots, setp
from numpy import arccos, ndarray, mod, argmin, degrees, linspace, array, sqrt, sin, cos

from .knots import create_knots
from .newton import xyz_newton_v, ta_newton_v
from .utils import mean_anomaly_offset, TWO_PI, eccentricity_vector
from .xyz5 import solve_xyz_o5s, xyz_o5v, cos_alpha_o5v, light_travel_time_o5v, vxyz_o5v, true_anomaly_o5v, rv_o5v, \
    star_planet_distance_o5v, ev_signal_o5v


class Orbit:
    def __init__(self, npt: int = 14, knot_placement: str = 'ea'):
        self.npt: int = npt
        self.times: Optional[ndarray] = None

        self._dt: Optional[float] = None
        self._points: Optional[float] = None
        self._coeffs: Optional[ndarray] = None
        self._t0: Optional[float] = None
        self._p: Optional[float] = None
        self._a: Optional[float] = None
        self._i: Optional[float] = None
        self._e: Optional[float] = None
        self._w: Optional[float] = None

        self._points, self._change_times, self._dt, self._tptable = create_knots(npt, 0.2, knot_placement)

    def set_data(self, times):
        self.times = times

    def set_pars(self, t0, p, a, i, e, w):
        self._t0 = t0
        self._p = p
        self._a = a
        self._i = i
        self._e = e
        self._w = w
        self._tc = t0 - mean_anomaly_offset(e, w)/TWO_PI*p
        self._coeffs = solve_xyz_o5s(self._points, p, a, i, e, w, self.npt)

    def mean_anomaly(self):
        offset = mean_anomaly_offset(self._e, self._w)
        return mod(TWO_PI * (self.times - (self._t0 - offset * self._p / TWO_PI)) / self._p, TWO_PI)

    def true_anomaly(self):
        ev = eccentricity_vector(self._i, self._e, self._w)
        return true_anomaly_o5v(self.times, self._t0, self._p, ev[0], ev[1], ev[2], self._w, self._dt, self._points, self._coeffs)

    def xyz(self, times: Optional[ndarray] = None):
        times = times if times is not None else self.times
        return xyz_o5v(times, self._tc, self._p, self._dt, self._tptable, self._points, self._coeffs)

    def _xyz_error(self):
        x, y, z = self.xyz()
        xt, yt, zt = xyz_newton_v(self.times, self._t0, self._p, self._a, self._i, self._e, self._w)
        return x-xt, y-yt, z-zt

    def vxyz(self):
        return vxyz_o5v(self.times, self._t0, self._p, self._dt, self._points, self._coeffs)

    def cos_phase(self):
        return cos_alpha_o5v(self.times, self._t0, self._p, self._dt, self._points, self._coeffs)

    def _cos_phase_error(self):
        ta = ta_newton_v(self.times, self._t0, self._p, self._e, self._w)
        cos_alpha_t = ta
        return cos_alpha_o5v(self.times, self._t0, self._p, self._dt, self._points, self._coeffs) - cos_alpha_t

    def phase(self):
        return arccos(cos_alpha_o5v(self.times, self._t0, self._p, self._dt, self._points, self._coeffs))

    def theta(self):
        return arccos(-cos_alpha_o5v(self.times, self._t0, self._p, self._dt, self._points, self._coeffs))

    def star_planet_distance(self, times: Optional[ndarray] = None):
        return star_planet_distance_o5v(times or self.times, self._t0, self._p, self._dt, self._points, self._coeffs)

    def light_travel_time(self, rstar: float):
        return light_travel_time_o5v(self.times, self._t0, self._p, rstar, self._dt, self._points, self._coeffs)

    def radial_velocity(self, k: float):
        return rv_o5v(self.times, k, self._t0, self._p, self._a, self._i, self._e, self._dt, self._points, self._coeffs)

    def ellipsoidal_variation(self, alpha: float, mass_ratio: float, times: Optional[ndarray] = None):
        """Ellipsoidal variation signal.

        NOTES: See Eqs. 6-10 in Lillo-Box al. (2014).
        """
        return ev_signal_o5v(alpha, mass_ratio, self._i, times or self.times, self._t0, self._p, self._dt, self._points, self._coeffs)

    def plot(self, figsize: Optional[tuple] = None, show_exact: bool = False, sr: float = 1.0, pr: float = 0.5, pc='k', npt: int = 1000):
        tcur = self.times
        self.set_data(linspace(0, self._p, npt))

        x, y, z = self.xyz()
        xl, yl, zl = 1.1*abs(x).max(), 1.1*abs(y).max(), 1.1*abs(z).max()
        al = max([xl, yl, zl])

        fig, axs = subplots(1, 3, figsize=figsize)
        axs[0].plot(x, y, zorder=0)
        axs[0].add_patch(Circle((self._coeffs[0, 0], self._coeffs[0, 1]), pr, fc=pc, ec='k', zorder=10))
        axs[1].plot(x, z, zorder=1)
        axs[1].add_patch(Circle((self._coeffs[0, 0], self._coeffs[0, 2]), pr, fc=pc, ec='k', zorder=11))
        axs[1].add_patch(Wedge((0, 0), 1.3*sr, 180 - degrees(self._w), 180, fc=pc, ec='k', zorder=-10))

        axs[2].plot(z, y, zorder=2)
        axs[2].add_patch(Circle((self._coeffs[0, 2], self._coeffs[0, 1]), pr, fc=pc, ec='k', zorder=12))

        di = self.times.size//6
        for i in range(6):
            axs[1].arrow(x[i*di], z[i*di], x[i*di + 1] - x[i*di], z[i*di + 1] - z[i*di], shape='full', lw=6,
                         length_includes_head=True, head_width=.1, color='k')

        m = x < 0.0
        axs[1].plot((0, x[m][argmin(abs(z[m]))]), (0, 0), 'k', zorder=-10, ls='--')
        omega_ix = argmin(x**2 + y**2 + z**2)
        axs[1].plot((0, x[omega_ix]), (0, z[omega_ix]), 'k', zorder=-10, ls='--')

        if show_exact:
            xt, yt, zt = xyz_newton_v(self.times, self._t0, self._p, self._a, self._i, self._e, self._w)
            axs[0].plot(xt, yt, 'k--')
            axs[1].plot(xt, zt, 'k--')
            axs[2].plot(zt, yt, 'k--')

        [ax.add_patch(Circle((0, 0), sr, fc='y', ec='k')) for ax in axs]
        [ax.set_aspect(1) for ax in axs]
        setp(axs, xlim=(-al, al), ylim=(-al, al))
        setp(axs[0], xlabel='X', ylabel='Y', title='Front')
        setp(axs[1], xlabel='X', ylabel='Z', ylim=(al, -al), title='Top')
        setp(axs[2], xlabel='Z', ylabel='Y', title='Side')
        fig.tight_layout()
        self.set_data(tcur)



#  ExOrbit: fast orbit calculations for exoplanet modelling
#  Copyright (C) 2022 Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit
from numpy import cos, sin, floor, sqrt, zeros, linspace, arccos, pi

from .newton import ta_newton_s
from .utils import mean_anomaly, mean_anomaly_offset


@njit(fastmath=True)
def solve_xyz_p5s(phase, p, a, i, e, w):
    """Planet velocity, acceleration, jerk, and snap at mid-transit in [R_star / day]"""

    # Time step for central finite difference
    # ---------------------------------------
    # I've tried to choose a value that is small enough to
    # work with ultra-short-period orbits and large enough
    # not to cause floating point problems with the fourth
    # derivative (anything much smaller starts hitting the
    # double precision limit.)
    dt = 2e-2

    ae = a*(1. - e**2)
    ci = cos(i)
    si = sin(i)

    # Calculation of X and Y positions
    # --------------------------------
    # These could just as well be calculated with a single
    # loop with X and Y as arrays, but I've decided to
    # manually unroll it because it seems to give a small
    # speed advantage with numba.

    f0 = ta_newton_s(phase-3 * dt, 0.0, p, e, w)
    f1 = ta_newton_s(phase-2 * dt, 0.0, p, e, w)
    f2 = ta_newton_s(phase-dt, 0.0, p, e, w)
    f3 = ta_newton_s(phase, 0.0, p, e, w)
    f4 = ta_newton_s(phase+dt, 0.0, p, e, w)
    f5 = ta_newton_s(phase+2 * dt, 0.0, p, e, w)
    f6 = ta_newton_s(phase+3 * dt, 0.0, p, e, w)

    r0 = ae/(1. + e*cos(f0))
    r1 = ae/(1. + e*cos(f1))
    r2 = ae/(1. + e*cos(f2))
    r3 = ae/(1. + e*cos(f3))
    r4 = ae/(1. + e*cos(f4))
    r5 = ae/(1. + e*cos(f5))
    r6 = ae/(1. + e*cos(f6))

    x0 = -r0*cos(w + f0)
    x1 = -r1*cos(w + f1)
    x2 = -r2*cos(w + f2)
    x3 = -r3*cos(w + f3)
    x4 = -r4*cos(w + f4)
    x5 = -r5*cos(w + f5)
    x6 = -r6*cos(w + f6)

    y0 = -r0*sin(w + f0)*ci
    y1 = -r1*sin(w + f1)*ci
    y2 = -r2*sin(w + f2)*ci
    y3 = -r3*sin(w + f3)*ci
    y4 = -r4*sin(w + f4)*ci
    y5 = -r5*sin(w + f5)*ci
    y6 = -r6*sin(w + f6)*ci

    z0 = r0*sin(w + f0)*si
    z1 = r1*sin(w + f1)*si
    z2 = r2*sin(w + f2)*si
    z3 = r3*sin(w + f3)*si
    z4 = r4*sin(w + f4)*si
    z5 = r5*sin(w + f5)*si
    z6 = r6*sin(w + f6)*si

    # First time derivative of position: velocity
    # -------------------------------------------
    a, b, c = 1/60, 9/60, 45/60
    vx = (a*(x6 - x0) + b*(x1 - x5) + c*(x4 - x2))/dt
    vy = (a*(y6 - y0) + b*(y1 - y5) + c*(y4 - y2))/dt
    vz = (a*(z6 - z0) + b*(z1 - z5) + c*(z4 - z2))/dt

    # Second time derivative of position: acceleration
    # ------------------------------------------------
    a, b, c, d = 1/90, 3/20, 3/2, 49/18
    ax = (a*(x0 + x6) - b*(x1 + x5) + c*(x2 + x4) - d*x3)/dt**2
    ay = (a*(y0 + y6) - b*(y1 + y5) + c*(y2 + y4) - d*y3)/dt**2
    az = (a*(z0 + z6) - b*(z1 + z5) + c*(z2 + z4) - d*z3)/dt**2

    # Third time derivative of position: jerk
    # ---------------------------------------
    a, b, c = 1/8, 1, 13/8
    jx = (a*(x0 - x6) + b*(x5 - x1) + c*(x2 - x4))/dt**3
    jy = (a*(y0 - y6) + b*(y5 - y1) + c*(y2 - y4))/dt**3
    jz = (a*(z0 - z6) + b*(z5 - z1) + c*(z2 - z4))/dt**3

    # Fourth time derivative of position: snap
    # ----------------------------------------
    a, b, c, d = 1/6, 2, 13/2, 28/3
    sx = (-a*(x0 + x6) + b*(x1 + x5) - c*(x2 + x4) + d*x3)/dt**4
    sy = (-a*(y0 + y6) + b*(y1 + y5) - c*(y2 + y4) + d*y3)/dt**4
    sz = (-a*(z0 + z6) + b*(z1 + z5) - c*(z2 + z4) + d*z3)/dt**4

    return x3, y3, z3, vx, vy, vz, ax, ay, az, jx, jy, jz, sx, sy, sz


@njit
def solve_xyz_o5s(knot_times, p, a, i, e, w, npt):
    coeffs = zeros((npt, 15))
    to = mean_anomaly_offset(e, w)/(2*pi)*p
    for ix in range(npt-1):
        coeffs[ix] = solve_xyz_p5s(p*knot_times[ix] - to, p, a, i, e, w)
    coeffs[-1] = coeffs[0]
    return coeffs


@njit(fastmath=True)
def xyz_o5s(t, t0, p, dt, pktable, points, cf):
    """Calculate planet's (x, y, z) position for a scalar time for any orbital phase"""
    epoch = floor((t - t0) / p)
    tc = t - t0 - epoch * p
    ix = pktable[int(floor(tc / (dt*p)))]
    #cf[ix] = x0, y0, z0, vx, vy, vz, ax, ay, az, jx, jy, jz, sx, sy, sz
    tc -= points[ix] * p
    tc2 = tc * tc
    tc3 = tc2 * tc
    tc4 = tc3 * tc
    px = cf[ix, 0] + cf[ix, 3] * tc + 0.5 * cf[ix, 6] * tc2 + cf[ix, 9]  * tc3 / 6.0 + cf[ix, 12] * tc4 / 24.
    py = cf[ix, 1] + cf[ix, 4] * tc + 0.5 * cf[ix, 7] * tc2 + cf[ix, 10] * tc3 / 6.0 + cf[ix, 13] * tc4 / 24.
    pz = cf[ix, 2] + cf[ix, 5] * tc + 0.5 * cf[ix, 8] * tc2 + cf[ix, 11] * tc3 / 6.0 + cf[ix, 14] * tc4 / 24.
    return px, py, pz


@njit(fastmath=True)
def xyz_o5v(times, t0, p, dt, pktable, points, coeffs):
    """Calculate planet's (x, y, z) position for a vector time for any orbital phase"""
    npt = times.size
    xs, ys, zs = zeros(npt), zeros(npt), zeros(npt)
    for i in range(npt):
        xs[i], ys[i], zs[i] = xyz_o5s(times[i], t0, p, dt, pktable, points, coeffs)
    return xs, ys, zs


@njit(fastmath=True)
def vxyz_o5s(t, t0, p, dt, points, cf):
    """Calculate planet's (x, y, z) velocity for a scalar time for any orbital phase"""
    epoch = floor((t - t0) / p)
    tc = t - t0 - epoch * p
    ix = int(floor(tc / dt + 0.5))
    tc -= points[ix]
    tc2 = tc * tc
    tc3 = tc2 * tc
    vx = cf[ix, 3] + cf[ix, 6] * tc + 0.5 * cf[ix, 9] * tc2 + cf[ix, 12]  * tc3 / 6.0
    vy = cf[ix, 4] + cf[ix, 7] * tc + 0.5 * cf[ix, 10] * tc2 + cf[ix, 13] * tc3 / 6.0
    vz = cf[ix, 5] + cf[ix, 8] * tc + 0.5 * cf[ix, 11] * tc2 + cf[ix, 14] * tc3 / 6.0
    return vx, vy, vz


@njit(fastmath=True)
def vxyz_o5v(times, t0, p, dt, points, coeffs):
    """Calculate planet's (x, y, z) position for a vector time for any orbital phase"""
    npt = times.size
    xs, ys, zs = zeros(npt), zeros(npt), zeros(npt)
    for i in range(npt):
        xs[i], ys[i], zs[i] = vxyz_o5s(times[i], t0, p, dt, points, coeffs)
    return xs, ys, zs


@njit(fastmath=True)
def vz_o5s(t, t0, p, dt, points, cf):
    """Calculate planet's (x, y, z) velocity for a scalar time for any orbital phase"""
    epoch = floor((t - t0) / p)
    tc = t - t0 - epoch * p
    ix = int(floor(tc / dt + 0.5))
    tc -= points[ix]
    tc2 = tc * tc
    tc3 = tc2 * tc
    return cf[ix, 5] + cf[ix, 8] * tc + 0.5 * cf[ix, 11] * tc2 + cf[ix, 14] * tc3 / 6.0


@njit(fastmath=True)
def vz_o5v(times, t0, p, dt, points, coeffs):
    """Calculate planet's (x, y, z) position for a vector time for any orbital phase"""
    npt = times.size
    vzs = zeros(npt)
    for i in range(npt):
        vzs[i] = vz_o5s(times[i], t0, p, dt, points, coeffs)
    return vzs


@njit(fastmath=True)
def pd_o5s(t, t0, p, dt, points, cf):
    """Calculate the projected planet-star center distance for a scalar time for any orbital phase"""
    epoch = floor((t - t0) / p)
    tc = t - t0 - epoch * p
    ix = int(floor(tc / dt + 0.5))
    tc -= points[ix]
    tc2 = tc * tc
    tc3 = tc2 * tc
    tc4 = tc3 * tc
    px = cf[ix, 0] + cf[ix, 3] * tc + 0.5 * cf[ix, 6] * tc2 + cf[ix, 9]  * tc3 / 6.0 + cf[ix, 12] * tc4 / 24.
    py = cf[ix, 1] + cf[ix, 4] * tc + 0.5 * cf[ix, 7] * tc2 + cf[ix, 10] * tc3 / 6.0 + cf[ix, 13] * tc4 / 24.
    return sqrt(px**2 + py**2)


@njit(fastmath=True)
def z_o5s(t, t0, p, dt, points, cf):
    """Calculate planet's (z) position for a scalar time for any orbital phase"""
    epoch = floor((t - t0) / p)
    tc = t - t0 - epoch * p
    ix = int(floor(tc / dt + 0.5))
    tc -= points[ix]
    tc2 = tc * tc
    tc3 = tc2 * tc
    tc4 = tc3 * tc
    pz = cf[ix, 2] + cf[ix, 5] * tc + 0.5 * cf[ix, 8] * tc2 + cf[ix, 11] * tc3 / 6.0 + cf[ix, 14] * tc4 / 24.
    return pz


@njit(fastmath=True)
def z_o5v(times, t0, p, dt, points, coeffs):
    """Calculate planet's (z) position for a vector time for any orbital phase"""
    npt = times.size
    zs = zeros(npt)
    for i in range(npt):
        zs[i] = z_o5s(times[i], t0, p, dt, points, coeffs)
    return zs


@njit(fastmath=True)
def xyz_t15s(tc, t0, p, x0, y0, z0, vx, vy, vz, ax, ay, az, jx, jy, jz, sx, sy, sz):
    """Calculate planet's (x,y) position near transit."""
    epoch = floor((tc - t0 + 0.5 * p) / p)
    t = tc - (t0 + epoch * p)
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    px = x0 + vx * t + 0.5 * ax * t2 + jx * t3 / 6.0 + sx * t4 / 24.
    py = y0 + vy * t + 0.5 * ay * t2 + jy * t3 / 6.0 + sy * t4 / 24.
    pz = z0 + vz * t + 0.5 * az * t2 + jz * t3 / 6.0 + sz * t4 / 24.
    return px, py, pz


@njit
def true_anomaly_o5v(times, t0, p, ex, ey, ez, w, dt, points, coeffs):
    npt = times.size
    f = zeros(npt)
    if ex <= -0.9999:
        f[:] = mean_anomaly(times, t0, p, 0.0, w)
    else:
        nes = (ex**2 + ey**2 + ez**2)
        for i in range(npt):
            x, y, z = xyz_o5s(times[i], t0, p, dt, points, coeffs)
            vx, vy, vz = vxyz_o5s(times[i], t0, p, dt, points, coeffs)
            edp = (x*ex + y*ey + z*ez) / sqrt((x**2 + y**2 + z**2) * nes)

            if edp <= -1.0:
                f[i] = pi
            elif edp >= 1.0:
                f[i] = 0.0
            elif (x*vx + y*vy + z*vz) > 0.0:
                f[i] = arccos(edp)
            else:
                f[i] = 2.0*pi - arccos(edp)
    return f


@njit
def cos_v_p_angle_o5v(v, times, t0, p, dt, points, coeffs):
    px, py, pz = xyz_o5v(times, t0, p, dt, points, coeffs)
    np = sqrt(px**2 + py**2 + pz**2)
    nv = sqrt(v[0]**2 + v[1]**2 + v[2]**2)
    return (px*v[0] + py*v[1] + pz*v[2])/(np*nv)


@njit
def cos_alpha_o5s(t, t0, p, dt, points, coeffs):
    """Cosine of the phase angle."""
    x, y, z = xyz_o5s(t, t0, p, dt, points, coeffs)
    return -z / sqrt(x**2 + y**2 + z**2)


@njit
def cos_alpha_o5v(times, t0, p, dt, points, coeffs):
    """Cosine of the phase angle."""
    x, y, z = xyz_o5v(times, t0, p, dt, points, coeffs)
    return -z / sqrt(x**2 + y**2 + z**2)


@njit
def star_planet_distance_o5v(times, t0, p, dt, points, coeffs):
    x, y, z = xyz_o5v(times, t0, p, dt, points, coeffs)
    return sqrt(x**2 + y**2 + z**2)


@njit
def ev_signal_o5v(alpha, mass_ratio, inc, times, t0, p, dt, points, coeffs):
    """Ellipsoidal variation signal.

    NOTES: See Eqs. 6-10 in Lillo-Box al. (2014).
    """
    x, y, z = xyz_o5v(times, t0, p, dt, points, coeffs)
    distance = sqrt(x**2 + y**2 + z**2)
    theta = arccos(z / distance)
    return -alpha * mass_ratio * sin(inc)**2 * cos(2*theta) / distance**3


@njit
def zdiff_o5s(t, t0, p, dt, points, coeffs):
    return z_o5s(t, t0, p, dt, points, coeffs) - coeffs[0, 2]


@njit
def light_travel_time_o5s(t, t0, p, rstar, dt, points, coeffs):
    """Light travel time in days."""
    s = 2.685885891543453e-05  # ((1 * u.R_sun).to(u.m) / c.c).to('d').value
    return -zdiff_o5s(t, t0, p, dt, points, coeffs) * rstar * s


@njit
def light_travel_time_o5v(times, t0, p, rstar, dt, points, coeffs):
    """Light travel time in days."""
    s = 2.685885891543453e-05  # ((1 * u.R_sun).to(u.m) / c.c).to('d').value
    ltt = zeros(times.size)
    for i in range(times.size):
        ltt[i] = -zdiff_o5s(times[i], t0, p, dt, points, coeffs) * rstar * s
    return ltt


@njit
def rv_o5s(time, k, t0, p, a, i, e, dt, points, coeffs):
    n = 2*pi/p * (a*sin(i))/sqrt(1-e**2)  # Perryman (2018) Eq. 2.23
    return vz_o5s(time, t0, p, dt, points, coeffs) / n * k


@njit
def rv_o5v(times, k, t0, p, a, i, e, dt, points, coeffs):
    npt = times.size
    rvs = zeros(npt)
    n = 2*pi/p * (a*sin(i))/sqrt(1-e**2)  # Perryman (2018) Eq. 2.23
    for i in range(npt):
        rvs[i] = vz_o5s(times[i], t0, p, dt, points, coeffs) / n * k
    return rvs


#  ExOrbit: fast orbit calculations for exoplanet modelling
#  Copyright (C) 2022 Hannu Parviainen
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

from numba import njit, types
from numpy import cos, sin, floor, sqrt, zeros, linspace, array

from ..newton import ta_newton_s


@njit(fastmath=True)
def solve_xy_p5s(phase, p, a, i, e, w):
    """Planet velocity, acceleration, jerk, and snap at mid-transit in [R_star / day]"""

    # Time step for central finite difference
    # ---------------------------------------
    # I've tried to choose a value that is small enough to
    # work with ultra-short-period orbits and large enough
    # not to cause floating point problems with the fourth
    # derivative (anything much smaller starts hitting the
    # double precision limit.)
    dt = 2e-2

    ae = a*(1. - e**2)
    ci = cos(i)

    # Calculation of X and Y positions
    # --------------------------------
    # These could just as well be calculated with a single
    # loop with X and Y as arrays, but I've decided to
    # manually unroll it because it seems to give a small
    # speed advantage with numba.

    f0 = ta_newton_s(phase-3 * dt, 0.0, p, e, w)
    f1 = ta_newton_s(phase-2 * dt, 0.0, p, e, w)
    f2 = ta_newton_s(phase-dt, 0.0, p, e, w)
    f3 = ta_newton_s(phase, 0.0, p, e, w)
    f4 = ta_newton_s(phase+dt, 0.0, p, e, w)
    f5 = ta_newton_s(phase+2 * dt, 0.0, p, e, w)
    f6 = ta_newton_s(phase+3 * dt, 0.0, p, e, w)

    r0 = ae/(1. + e*cos(f0))
    r1 = ae/(1. + e*cos(f1))
    r2 = ae/(1. + e*cos(f2))
    r3 = ae/(1. + e*cos(f3))
    r4 = ae/(1. + e*cos(f4))
    r5 = ae/(1. + e*cos(f5))
    r6 = ae/(1. + e*cos(f6))

    x0 = -r0*cos(w + f0)
    x1 = -r1*cos(w + f1)
    x2 = -r2*cos(w + f2)
    x3 = -r3*cos(w + f3)
    x4 = -r4*cos(w + f4)
    x5 = -r5*cos(w + f5)
    x6 = -r6*cos(w + f6)

    y0 = -r0*sin(w + f0)*ci
    y1 = -r1*sin(w + f1)*ci
    y2 = -r2*sin(w + f2)*ci
    y3 = -r3*sin(w + f3)*ci
    y4 = -r4*sin(w + f4)*ci
    y5 = -r5*sin(w + f5)*ci
    y6 = -r6*sin(w + f6)*ci

    cf = zeros((2, 5))

    cf[0, 0] = x3
    cf[1, 0] = y3

    # First time derivative of position: velocity
    # -------------------------------------------
    a, b, c = 1/60, 9/60, 45/60
    cf[0, 1] = (a*(x6 - x0) + b*(x1 - x5) + c*(x4 - x2))/dt  # vx
    cf[1, 1] = (a*(y6 - y0) + b*(y1 - y5) + c*(y4 - y2))/dt  # vy

    # Second time derivative of position: acceleration
    # ------------------------------------------------
    a, b, c, d = 1/90, 3/20, 3/2, 49/18
    cf[0, 2] = (a*(x0 + x6) - b*(x1 + x5) + c*(x2 + x4) - d*x3)/dt**2  # ax
    cf[1, 2] = (a*(y0 + y6) - b*(y1 + y5) + c*(y2 + y4) - d*y3)/dt**2  # ay

    # Third time derivative of position: jerk
    # ---------------------------------------
    a, b, c = 1/8, 1, 13/8
    cf[0, 3] = (a*(x0 - x6) + b*(x5 - x1) + c*(x2 - x4))/dt**3
    cf[1, 3] = (a*(y0 - y6) + b*(y5 - y1) + c*(y2 - y4))/dt**3

    # Fourth time derivative of position: snap
    # ----------------------------------------
    a, b, c, d = 1/6, 2, 13/2, 28/3
    cf[0, 4] = (-a*(x0 + x6) + b*(x1 + x5) - c*(x2 + x4) + d*x3)/dt**4
    cf[1, 4] = (-a*(y0 + y6) + b*(y1 + y5) - c*(y2 + y4) + d*y3)/dt**4

    return cf


@njit
def solve_xy_t25(dt, p, a, i, e, w):
    c1 = array(solve_xy_p5s(-0.5*dt, p, a, i, e, w))
    c2 = array(solve_xy_p5s(0.5*dt, p, a, i, e, w))
    return c1, c2


@njit
def solve_xy_o5s(p, a, i, e, w, npt):
    points = linspace(0.0, p, npt)
    dt = points[1] - points[0]
    coeffs = zeros((npt, 10))
    for ix in range(npt-1):
        coeffs[ix] = solve_xy_p5s(points[ix], p, a, i, e, w)
    coeffs[-1] = coeffs[0]
    return dt, points, coeffs


@njit(fastmath=True)
def xy_t15s(tc, t0, p, c):
    """Calculate planet's (x,y) position near transit."""
    epoch = floor((tc - t0 + 0.5 * p) / p)
    t = tc - (t0 + epoch * p)
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    px = c[0, 0] + c[0, 1] * t + 0.5 * c[0, 2] * t2 + c[0, 3] * t3 / 6.0 + c[0, 4] * t4 / 24.
    py = c[1, 0] + c[1, 1] * t + 0.5 * c[1, 2] * t2 + c[1, 3] * t3 / 6.0 + c[1, 4] * t4 / 24.
    return px, py


@njit(fastmath=True)
def xy_t15sc(t, c):
    """Calculate planet's (x,y) position near transit."""
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    px = c[0, 0] + c[0, 1] * t + 0.5 * c[0, 2] * t2 + c[0, 3] * t3 / 6.0 + c[0, 4] * t4 / 24.
    py = c[1, 0] + c[1, 1] * t + 0.5 * c[1, 2] * t2 + c[1, 3] * t3 / 6.0 + c[1, 4] * t4 / 24.
    return px, py


@njit(fastmath=True)
def xyd_t15s(t, c):
    """Calculate planet's (x,y) position near transit."""
    t2 = t * t
    t3 = t2 * t
    t4 = t3 * t
    px = c[0, 0] + c[0, 1] * t + 0.5 * c[0, 2] * t2 + c[0, 3] * t3 / 6.0 + c[0, 4] * t4 / 24.
    py = c[1, 0] + c[1, 1] * t + 0.5 * c[1, 2] * t2 + c[1, 3] * t3 / 6.0 + c[1, 4] * t4 / 24.
    return px, py, sqrt(px**2 + py**2)


@njit
def xy_t15v(tc, t0, p, c):
    npt = tc.size
    xs, ys = zeros(npt), zeros(npt)
    for i in range(npt):
        xs[i], ys[i] = xy_t15s(tc[i], t0, p, c)
    return xs, ys


def xy_t15(tc, t0, p, c):
    if isinstance(tc, types.Float):
        return xy_t15s(tc, t0, p, c)
    else:
        return xy_t15v(tc, t0, p, c)


@njit(fastmath=True)
def pd_t15(tc, t0, p, c):
    """Calculate the (p)rojected planet-star center (d)istance near (t)ransit."""
    px, py = xy_t15(tc, t0, p, c)
    return sqrt(px ** 2 + py ** 2)


@njit(fastmath=True)
def pd_t15sc(tc, c):
    """Calculate the (p)rojected planet-star center (d)istance near (t)ransit."""
    px, py = xy_t15sc(tc, c)
    return sqrt(px ** 2 + py ** 2)


@njit(fastmath=True)
def pd_t25s(t, t0, p, dt, c1, c2):
    """Slower but more accurate (p)rojected planet-star center (d)istance near (t)ransit for scalar time.

    A more accurate version of planet-star center distance calculation that interpolates between two Taylor
    series expansions around the transit center. Much slower than `pd_t15s` and you're unlikely really going
    to need the added precision. Use `solve_xy_t25d` to compute the coefficient arrays.
    """
    epoch = floor((t - t0 + 0.5 * p) / p)
    tg = t - (t0 + epoch * p)
    dt = 0.5*dt

    if tg < dt:
        t1 = tg + dt
        t2 = t1 * t1
        t3 = t2 * t1
        t4 = t3 * t1
        px1 = c1[0] + c1[2] * t1 + 0.5 * c1[4] * t2 + c1[6] * t3 / 6.0 + c1[8] * t4 / 24.
        py1 = c1[1] + c1[3] * t1 + 0.5 * c1[5] * t2 + c1[7] * t3 / 6.0 + c1[9] * t4 / 24.
    else:
        px1, py1 = 0.0, 0.0

    if tg > -dt:
        t1 = tg - dt
        t2 = t1 * t1
        t3 = t2 * t1
        t4 = t3 * t1
        px2 = c2[0] + c2[2] * t1 + 0.5 * c2[4] * t2 + c2[6] * t3 / 6.0 + c2[8] * t4 / 24.
        py2 = c2[1] + c2[3] * t1 + 0.5 * c2[5] * t2 + c2[7] * t3 / 6.0 + c2[9] * t4 / 24.
    else:
        px2, py2 = 0.0, 0.0

    if tg < -dt:
        return sqrt(px1 ** 2 + py1 ** 2)
    elif tg > dt:
        return sqrt(px2 ** 2 + py2 ** 2)
    else:
        a = (tg + dt) / (2 * dt)
        px = (1 - a) * px1 + a * px2
        py = (1 - a) * py1 + a * py2
        return sqrt(px ** 2 + py ** 2)


@njit(fastmath=True)
def pd_t25v(times, t0, p, dt, c1, c2):
    """Slower but more accurate (p)rojected planet-star center (d)istance near (t)ransit for a time array.

    A more accurate version of planet-star center distance calculation that interpolates between two Taylor
    series expansions around the transit center. Much slower than `pd_t15s` and you're unlikely really going
    to need the added precision. Use `solve_xy_t25d` to compute the coefficient arrays.
    """
    z = zeros(times.size)
    for i in range(times.size):
        z[i] = pd_t25s(times[i], t0, p, dt, c1, c2)
    return z



from math import sin, sqrt, pi, cos

from numba import njit
from numpy import zeros, arctan2, array, atleast_1d, asarray
from scipy.constants import G

from ..utils import as_from_rhop, i_from_baew
from .position import solve_xy_p5s
from .derivatives import xy_derivative_coeffs, loc_and_der_coeffs


@njit
def pds_diff(p, rho, b, secw, sesw, dd):
    v = zeros((6, 6))
    a0 = as_from_rhop(rho, p)
    e0 = secw ** 2 + sesw ** 2
    w0 = arctan2(sesw, secw)
    i0 = i_from_baew(b, a0, e0, w0)

    v[:, :] = array([0.0, p, a0, i0, e0, w0])
    v[0, 0] -= dd

    # Period: p, a, and i
    a1 = as_from_rhop(rho, p + dd)
    i1 = i_from_baew(b, a1, e0, w0)
    v[1, 1] += dd
    v[1, 2] = a1
    v[1, 3] = i1

    # Stellar density: a and i
    a1 = as_from_rhop(rho + dd, p)
    i1 = i_from_baew(b, a1, e0, w0)
    v[2, 2] = a1
    v[2, 3] = i1

    # Impact parameter: i
    i1 = i_from_baew(b + dd, a0, e0, w0)
    v[3, 3] = i1

    # sqrt e cos w: i, e, and w
    e1 = (secw + dd) ** 2 + sesw ** 2
    w1 = arctan2(sesw, secw + dd)
    i1 = i_from_baew(b, a0, e1, w1)
    v[4, 3] = i1
    v[4, 4] = e1
    v[4, 5] = w1

    # sqrt e sin w: i, e, and w
    e2 = secw ** 2 + (sesw + dd) ** 2
    w2 = arctan2(sesw + dd, secw)
    i2 = i_from_baew(b, a0, e2, w2)
    v[5, 3] = i2
    v[5, 4] = e2
    v[5, 5] = w2
    return v


@njit
def partial_derivatives(p, rho, b, secw, sesw):
    v = zeros((6, 6))
    a = as_from_rhop(rho, p)
    e = secw ** 2 + sesw ** 2
    w = arctan2(sesw, secw)
    i = i_from_baew(b, a, e, w)

    v[0, 0] = -1

    # Period: p, a, and i
    da = da_dp(rho, p)
    di = di_da(b, a, e, w)
    v[1, 1] = 1
    v[1, 2] = da
    v[1, 3] = da * di

    # Stellar density: a and i
    da = da_drho(rho, p)
    di = di_da(b, a, e, w)
    v[2, 2] = da
    v[2, 3] = da * di

    # Impact parameter: i
    v[3, 3] = di_db(b, a, e, w)

    # secw and sesw
    # -------------
    die = di_de(b, a, e, w)
    diw = di_dw(b, a, e, w)

    # sqrt e cos w: i, e, and w
    de = de_dsecw(secw)
    dw = dw_dsecw(secw, sesw)
    v[4, 3] = de * die + dw * diw
    v[4, 4] = de
    v[4, 5] = dw

    # sqrt e sin w: i, e, and w
    de = de_dsesw(sesw)
    dw = dw_dsesw(secw, sesw)
    v[5, 3] = de * die + dw * diw
    v[5, 4] = de
    v[5, 5] = dw
    return v


@njit
def coeffs_old(phase, p, rho, b, secw, sesw):
    a = as_from_rhop(rho, p)
    e = secw ** 2 + sesw ** 2
    w = arctan2(sesw, secw)
    i = i_from_baew(b, a, e, w)

    coeffs = solve_xy_p5s(phase, p, a, i, e, w)
    pds = partial_derivatives(p, rho, b, secw, sesw)
    dcoeffs = xy_derivative_coeffs(0.0, p, a, i, e, w, pds, coeffs, 1e-4)
    return coeffs, dcoeffs

@njit
def coeffs(phase, p, rho, b, secw, sesw, with_derivatives):
    p = atleast_1d(asarray(p))
    rho = atleast_1d(asarray(rho))
    b = atleast_1d(asarray(b))
    secw = atleast_1d(asarray(secw))
    sesw = atleast_1d(asarray(sesw))
    nor = p.size

    a = as_from_rhop(rho, p)
    e = secw ** 2 + sesw ** 2
    w = arctan2(sesw, secw)
    i = i_from_baew(b, a, e, w)

    if with_derivatives:
        diffs = zeros((nor, 6, 6))
        for j in range(nor):
            diffs[j] = partial_derivatives(p[j], rho[j], b[j], secw[j], sesw[j])
    else:
        diffs = None

    return loc_and_der_coeffs(phase, p, a, i, e, w, diffs)

# Partial derivatives
# -------------------

@njit
def da_dp(rho, p):
    return (2 / 3) * (G * (1e3 * rho * 86400 ** 2) / (3 * pi * p)) ** (1 / 3)


@njit
def da_drho(rho, p):
    return (1 / 3) * ((G * 86400 ** 2 * p ** 2 * 1e3) / (3 * pi)) ** (1 / 3) * rho ** (-2 / 3)


@njit
def di_da(b, a, e, w):
    ea = (1.0 - e ** 2) / (1.0 + e * sin(w))
    return (b / (a ** 2 * ea)) / sqrt(1.0 - (b / (a * ea)) ** 2)


@njit
def di_db(b, a, e, w):
    ea = (1.0 - e ** 2) / (1.0 + e * sin(w))
    return -1.0 / (a * ea * sqrt(1.0 - (b / (a * ea)) ** 2))


@njit
def di_de(b, a, e, w):
    l = b / a
    m = (e ** 2 - 1) ** 2
    return -l * (e ** 2 * sin(w) + 2 * e + sin(w)) / (sqrt(-(l ** 2 * (e * sin(w) + 1) ** 2 - m) / m) * m)


@njit
def di_dw(b, a, e, w):
    l = b / a
    m = (e ** 2 - 1) ** 2
    return l * e * cos(w) / (sqrt((-l ** 2 * (e * sin(w) + 1) ** 2 + m) / m) * (e ** 2 - 1))


@njit
def de_dsecw(secw):
    return 2 * secw


@njit
def de_dsesw(sesw):
    return 2 * sesw


@njit
def dw_dsecw(secw, sesw):
    return -sesw / (secw ** 2 + sesw ** 2)


@njit
def dw_dsesw(secw, sesw):
    return secw / (secw ** 2 + sesw ** 2)


from numba import njit
from numpy import pi, linspace, zeros
from scipy.optimize import root_scalar

from .newton import ea_newton_s, ea_newton_v, ta_newton_s, ta_newton_v, xyz_newton_v

@njit
def eccentric_anomaly(t, e):
    return ea_newton_s(t, 0.0, 1.0, e, 0.5*pi)

@njit
def true_anomaly(t, e):
    f = ta_newton_s(t, 0.0, 1.0, e, 0.5*pi)
    if f < 0.0:
        f += 2*pi
    return f


def create_knots(n_knots: int, e: float, quantity: str = 'ea', tres: int = 200):
    if quantity not in ('mm', 'ea', 'ta'):
        raise ValueError("Quantity needs to be either 'mm' for mean motion, 'ea' for eccentric anomaly, or 'ta' for true anomaly.")
    if n_knots % 2 != 1:
        raise ValueError("Number of knots should be odd.")

    if quantity == 'mm':
        knot_times = linspace(0, 1, n_knots++1, endpoint=False)
        dt = 1 / n_knots
        change_times = linspace(0.5 * dt, 1 - 0.5 * dt, n_knots)
    else:
        if quantity == 'ea':
            def cfun(t, e, v):
                return eccentric_anomaly(t, e) - v
        elif quantity == 'ta':
            def cfun(t, e, v):
                return true_anomaly(t, e) - v
        else:
            raise NotImplementedError

        knot_sep = 2 * pi / n_knots

        knot_times = zeros(n_knots)
        knot_times[n_knots // 2] = 0.5
        t0 = 1e-5
        for i in range(1, n_knots // 2):
            knot_times[i] = root_scalar(cfun, args=(e, i * knot_sep), bracket=(t0, 1.0 - 1e-5)).root
            t0 = knot_times[i]
        knot_times[n_knots // 2 + 1:-1] = 1 - knot_times[n_knots // 2 - 1:0:-1]
        knot_times[-1] = 1.0

        change_times = zeros(n_knots-1)
        t0 = 1e-5
        for i in range(0, n_knots // 2):
            change_times[i] = root_scalar(cfun, args=(e, (i + 0.5) * knot_sep), bracket=(t0, 1.0 - 1e-5)).root
            t0 = change_times[i]
        change_times[n_knots // 2:] = 1 - change_times[n_knots // 2 - 1::-1]

    # Create the time-to-knot table
    dt = 1 / tres
    tktable = zeros(tres, int)
    ik = 0
    for i in range(tres):
        if i*dt > change_times[ik]:
            ik += 1
        if ik >= n_knots-1:
            tktable[i:] = ik
            break
        tktable[i] = ik

    return knot_times, change_times, dt, tktable



from numpy import asarray, zeros, arctan2

from .utils import as_from_rhop, i_from_baew
from .xy.position import pd_t15, solve_xy_p5s, xy_t15
from .xy.derivatives import pd_with_derivatives_v, xy_derivative_coeffs
from .xy.par_direct import diffs as diffs_natural
from .xy.par_fitting import partial_derivatives as diffs_fitting

class Knot2D:

    def __init__(self, phase: float, t0: float, p: float, a: float, i: float, e: float, w: float,
                 derivatives: bool = False):
        self.derivatives = derivatives
        self.phase = phase
        self.t0 = t0
        self.p = p
        self.a = a
        self.i = i
        self.e = e
        self.w = w

        self._coeffs = solve_xy_p5s(phase, p, a, i, e, w)
        if derivatives:
            self._c_derivative_coeffs()

    def _c_derivative_coeffs(self):
        d = diffs_natural(self.p, self.a, self.i, self.e, self.w, 1e-4)
        self._coeffs_d = xy_derivative_coeffs(d, 1e-4, self._coeffs)

    def position(self, t):
        return xy_t15(t, self.t0, self.p, self._coeffs)

    def projected_distance(self, t):
        if self.derivatives:
            return pd_with_derivatives_v(t, self.t0, self.p, self._coeffs, self._coeffs_d)
        else:
            return pd_t15(t, self.t0, self.p, self._coeffs)

    def _pd_numerical_derivatives(self, t, e=1e-4):
        t = asarray(t)
        res = zeros((6, t.size))
        r0 = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase, self.p, self.a, self.i, self.e, self.w))
        res[0] = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase + e, self.p, self.a, self.i, self.e, self.w))
        res[1] = pd_t15(t, self.t0, self.p + e, solve_xy_p5s(self.phase, self.p + e, self.a, self.i, self.e, self.w))
        res[2] = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase, self.p, self.a + e, self.i, self.e, self.w))
        res[3] = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase, self.p, self.a, self.i + e, self.e, self.w))
        res[4] = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase, self.p, self.a, self.i, self.e + e, self.w))
        res[5] = pd_t15(t, self.t0, self.p, solve_xy_p5s(self.phase, self.p, self.a, self.i, self.e, self.w + e))
        res = (res - r0) / e
        return res


class Knot2DFit(Knot2D):

    def __init__(self, phase: float, t0: float, p: float, rho: float, b: float, secw: float, sesw: float,
                 derivatives: bool = False):
        self.rho = rho
        self.b = b
        self.secw = secw
        self.sesw = sesw
        a = as_from_rhop(rho, p)
        e = secw ** 2 + sesw ** 2
        w = arctan2(sesw, secw)
        i = i_from_baew(b, a, e, w)
        super().__init__(phase, t0, p, a, i, e, w, derivatives)

    def _c_derivative_coeffs(self):
        d = diffs_fitting(self.p, self.rho, self.b, self.secw, self.sesw, 1e-4)
        self._coeffs_d = xy_derivative_coeffs(d, 1e-4, self._coeffs)


from numba import njit
from numpy import zeros, array, atleast_1d, asarray

from .position import solve_xy_p5s
from .derivatives import xy_derivative_coeffs, loc_and_der_coeffs


@njit
def partial_derivatives():
    v = zeros((6, 6))
    v[0, 0] = -1.0
    for i in range(1, 6):
        v[i, i] = 1.0
    return v


@njit
def coeffs_old(phase, p, a, i, e, w):
    coeffs = solve_xy_p5s(phase, p, a, i, e, w)
    pds =  partial_derivatives()
    dcoeffs = xy_derivative_coeffs(0.0, p, a, i, e, w, pds, coeffs, 1e-4)
    return coeffs, dcoeffs

@njit
def coeffs(phase, p, a, i, e, w, with_derivatives):
    p = atleast_1d(asarray(p))
    a = atleast_1d(asarray(a))
    i = atleast_1d(asarray(i))
    e = atleast_1d(asarray(e))
    w = atleast_1d(asarray(w))
    nor = p.size

    if with_derivatives:
        diffs = zeros((nor, 6, 6))
        for j in range(nor):
            diffs[j] = partial_derivatives()
    else:
        diffs = None

    return loc_and_der_coeffs(phase, p, a, i, e, w, diffs)